from django.core import validators
from django.core.exceptions import ValidationError
from django.forms.widgets import Textarea

EMPTY_VALUES = validators.EMPTY_VALUES + ('[]',)


class CSVWidget(Textarea):
    is_hidden = False

    def prep_value(self, value):
        """Prepare value before effectively render widget"""
        if value in EMPTY_VALUES:
            return ""
        elif isinstance(value, str):
            return value
        elif isinstance(value, list):
            return ";".join(value)
        raise ValidationError('Invalid format.')

    def render(self, name, value, **kwargs):
        value = self.prep_value(value)
        return super().render(name, value, **kwargs)

    def value_from_datadict(self, data, files, name):
        """
        Return the value of this widget or None.

        Since we're only given the value of the entity name and the data dict
        contains the '_eav_config_cls' (which we don't have access to) as the
        key, we need to loop through each field checking if the eav attribute
        exists with the given 'name'.
        """
        widget_value = None
        for data_value in data:
            try:
                widget_value = getattr(data.get(data_value), name)
            except AttributeError:
                pass  # noqa: WPS420

        return widget_value



# Generated by Django 3.2 on 2021-05-10 13:05

from django.db import migrations

import eav.fields


class Migration(migrations.Migration):
    dependencies = [
        ('eav', '0004_alter_value_value_bool'),
    ]

    operations = [
        migrations.AddField(
            model_name='value',
            name='value_csv',
            field=eav.fields.CSVField(blank=True, default="", null=True),
        ),
        migrations.AlterField(
            model_name='attribute',
            name='datatype',
            field=eav.fields.EavDatatypeField(
                choices=[
                    ('text', 'Text'),
                    ('date', 'Date'),
                    ('float', 'Float'),
                    ('int', 'Integer'),
                    ('bool', 'True / False'),
                    ('object', 'Django Object'),
                    ('enum', 'Multiple Choice'),
                    ('json', 'JSON Object'),
                    ('csv', 'Comma-Separated-Value'),
                ],
                max_length=6,
                verbose_name='Data Type',
            ),
        ),
    ]



from django.db.models.fields import UUIDField


def get_entity_pk_type(entity_cls) -> str:
    """Returns the entity PK type to use.

    These values map to `models.Value` as potential fields to use to relate
    to the proper entity via the correct PK type.
    """
    if isinstance(entity_cls._meta.pk, UUIDField):
        return 'entity_uuid'
    return 'entity_id'



"""
This module defines the four concrete, non-abstract models:
    * :class:`Value`
    * :class:`Attribute`
    * :class:`EnumValue`
    * :class:`EnumGroup`

Along with the :class:`Entity` helper class and :class:`EAVModelMeta`
optional metaclass for each eav model class.
"""

from copy import copy

from django.contrib.contenttypes import fields as generic
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models.base import ModelBase
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from eav import register
from eav.exceptions import IllegalAssignmentException
from eav.fields import CSVField, EavDatatypeField
from eav.logic.entity_pk import get_entity_pk_type
from eav.logic.slug import SLUGFIELD_MAX_LENGTH, generate_slug
from eav.validators import (
    validate_bool,
    validate_csv,
    validate_date,
    validate_enum,
    validate_float,
    validate_int,
    validate_json,
    validate_object,
    validate_text,
)

try:
    from typing import Final
except ImportError:
    from typing_extensions import Final


CHARFIELD_LENGTH: Final = 100


class EnumValue(models.Model):
    """
    *EnumValue* objects are the value 'choices' to multiple choice *TYPE_ENUM*
    :class:`Attribute` objects. They have only one field, *value*, a
    ``CharField`` that must be unique.

    For example::

        yes = EnumValue.objects.create(value='Yes') # doctest: SKIP
        no = EnumValue.objects.create(value='No')
        unknown = EnumValue.objects.create(value='Unknown')

        ynu = EnumGroup.objects.create(name='Yes / No / Unknown')
        ynu.values.add(yes, no, unknown)

        Attribute.objects.create(name='has fever?',
            datatype=Attribute.TYPE_ENUM, enum_group=ynu)
        # = <Attribute: has fever? (Multiple Choice)>

    .. note::
       The same *EnumValue* objects should be reused within multiple
       *EnumGroups*.  For example, if you have one *EnumGroup* called: *Yes /
       No / Unknown* and another called *Yes / No / Not applicable*, you should
       only have a total of four *EnumValues* objects, as you should have used
       the same *Yes* and *No* *EnumValues* for both *EnumGroups*.
    """

    class Meta:
        verbose_name = _('EnumValue')
        verbose_name_plural = _('EnumValues')

    value = models.CharField(
        _('Value'),
        db_index=True,
        unique=True,
        max_length=SLUGFIELD_MAX_LENGTH,
    )

    def __str__(self):
        """String representation of `EnumValue` instance."""
        return str(self.value)

    def __repr__(self):
        """String representation of `EnumValue` object."""
        return '<EnumValue {0}>'.format(self.value)


class EnumGroup(models.Model):
    """
    *EnumGroup* objects have two fields - a *name* ``CharField`` and *values*,
    a ``ManyToManyField`` to :class:`EnumValue`. :class:`Attribute` classes
    with datatype *TYPE_ENUM* have a ``ForeignKey`` field to *EnumGroup*.

    See :class:`EnumValue` for an example.
    """

    class Meta:
        verbose_name = _('EnumGroup')
        verbose_name_plural = _('EnumGroups')

    name = models.CharField(
        unique=True,
        max_length=CHARFIELD_LENGTH,
        verbose_name=_('Name'),
    )
    values = models.ManyToManyField(
        EnumValue,
        verbose_name=_('Enum group'),
    )

    def __str__(self):
        """String representation of `EnumGroup` instance."""
        return str(self.name)

    def __repr__(self):
        """String representation of `EnumGroup` object."""
        return '<EnumGroup {0}>'.format(self.name)


class Attribute(models.Model):
    """
    Putting the **A** in *EAV*. This holds the attributes, or concepts.
    Examples of possible *Attributes*: color, height, weight, number of
    children, number of patients, has fever?, etc...

    Each attribute has a name, and a description, along with a slug that must
    be unique.  If you don't provide a slug, a default slug (derived from
    name), will be created.

    The *required* field is a boolean that indicates whether this EAV attribute
    is required for entities to which it applies. It defaults to *False*.

    .. warning::
       Just like a normal model field that is required, you will not be able
       to save or create any entity object for which this attribute applies,
       without first setting this EAV attribute.

    There are 7 possible values for datatype:

        * int (TYPE_INT)
        * float (TYPE_FLOAT)
        * text (TYPE_TEXT)
        * date (TYPE_DATE)
        * bool (TYPE_BOOLEAN)
        * object (TYPE_OBJECT)
        * enum (TYPE_ENUM)
        * json (TYPE_JSON)
        * csv (TYPE_CSV)


    Examples::

        Attribute.objects.create(name='Height', datatype=Attribute.TYPE_INT)
        # = <Attribute: Height (Integer)>

        Attribute.objects.create(name='Color', datatype=Attribute.TYPE_TEXT)
        # = <Attribute: Color (Text)>

        yes = EnumValue.objects.create(value='yes')
        no = EnumValue.objects.create(value='no')
        unknown = EnumValue.objects.create(value='unknown')
        ynu = EnumGroup.objects.create(name='Yes / No / Unknown')
        ynu.values.add(yes, no, unknown)

        Attribute.objects.create(name='has fever?', datatype=Attribute.TYPE_ENUM, enum_group=ynu)
        # = <Attribute: has fever? (Multiple Choice)>

    .. warning:: Once an Attribute has been used by an entity, you can not
                 change it's datatype.
    """

    class Meta:
        ordering = ['name']
        verbose_name = _('Attribute')
        verbose_name_plural = _('Attributes')

    TYPE_TEXT = 'text'
    TYPE_FLOAT = 'float'
    TYPE_INT = 'int'
    TYPE_DATE = 'date'
    TYPE_BOOLEAN = 'bool'
    TYPE_OBJECT = 'object'
    TYPE_ENUM = 'enum'
    TYPE_JSON = 'json'
    TYPE_CSV = 'csv'

    DATATYPE_CHOICES = (
        (TYPE_TEXT, _('Text')),
        (TYPE_DATE, _('Date')),
        (TYPE_FLOAT, _('Float')),
        (TYPE_INT, _('Integer')),
        (TYPE_BOOLEAN, _('True / False')),
        (TYPE_OBJECT, _('Django Object')),
        (TYPE_ENUM, _('Multiple Choice')),
        (TYPE_JSON, _('JSON Object')),
        (TYPE_CSV, _('Comma-Separated-Value')),
    )

    # Core attributes

    datatype = EavDatatypeField(
        choices=DATATYPE_CHOICES,
        max_length=6,
        verbose_name=_('Data Type'),
    )

    name = models.CharField(
        max_length=CHARFIELD_LENGTH,
        help_text=_('User-friendly attribute name'),
        verbose_name=_('Name'),
    )

    """
    Main identifer for the attribute.
    Upon creation, slug is autogenerated from the name.
    (see :meth:`~eav.fields.EavSlugField.create_slug_from_name`).
    """
    slug = models.SlugField(
        max_length=SLUGFIELD_MAX_LENGTH,
        db_index=True,
        unique=True,
        help_text=_('Short unique attribute label'),
        verbose_name=_('Slug'),
    )

    """
    .. warning::
        This attribute should be used with caution. Setting this to *True*
        means that *all* entities that *can* have this attribute will
        be required to have a value for it.
    """
    required = models.BooleanField(
        default=False,
        verbose_name=_('Required'),
    )

    entity_ct = models.ManyToManyField(
        ContentType,
        blank=True,
        verbose_name=_('Entity content type'),
    )
    """
    This field allows you to specify a relationship with any number of content types.
    This would be useful, for example, if you wanted an attribute to apply only to
    a subset of entities. In that case, you could filter by content type in the
    :meth:`~eav.registry.EavConfig.get_attributes` method of that entity's config.
    """

    enum_group = models.ForeignKey(
        EnumGroup,
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_('Choice Group'),
    )

    description = models.CharField(
        max_length=256,
        blank=True,
        null=True,
        help_text=_('Short description'),
        verbose_name=_('Description'),
    )

    # Useful meta-information

    display_order = models.PositiveIntegerField(
        default=1,
        verbose_name=_('Display order'),
    )

    modified = models.DateTimeField(
        auto_now=True,
        verbose_name=_('Modified'),
    )

    created = models.DateTimeField(
        default=timezone.now,
        editable=False,
        verbose_name=_('Created'),
    )

    @property
    def help_text(self):
        return self.description

    def get_validators(self):
        """
        Returns the appropriate validator function from :mod:`~eav.validators`
        as a list (of length one) for the datatype.

        .. note::
           The reason it returns it as a list, is eventually we may want this
           method to look elsewhere for additional attribute specific
           validators to return as well as the default, built-in one.
        """
        DATATYPE_VALIDATORS = {
            'text': validate_text,
            'float': validate_float,
            'int': validate_int,
            'date': validate_date,
            'bool': validate_bool,
            'object': validate_object,
            'enum': validate_enum,
            'json': validate_json,
            'csv': validate_csv,
        }

        return [DATATYPE_VALIDATORS[self.datatype]]

    def validate_value(self, value):
        """
        Check *value* against the validators returned by
        :meth:`get_validators` for this attribute.
        """
        for validator in self.get_validators():
            validator(value)

        if self.datatype == self.TYPE_ENUM:
            if isinstance(value, EnumValue):
                value = value.value
            if not self.enum_group.values.filter(value=value).exists():
                raise ValidationError(
                    _('%(val)s is not a valid choice for %(attr)s')
                    % dict(val=value, attr=self)
                )

    def save(self, *args, **kwargs):
        """
        Saves the Attribute and auto-generates a slug field
        if one wasn't provided.
        """
        if not self.slug:
            self.slug = generate_slug(self.name)

        self.full_clean()
        super(Attribute, self).save(*args, **kwargs)

    def clean(self):
        """
        Validates the attribute.  Will raise ``ValidationError`` if the
        attribute's datatype is *TYPE_ENUM* and enum_group is not set, or if
        the attribute is not *TYPE_ENUM* and the enum group is set.
        """
        if self.datatype == self.TYPE_ENUM and not self.enum_group:
            raise ValidationError(
                _('You must set the choice group for multiple choice attributes')
            )

        if self.datatype != self.TYPE_ENUM and self.enum_group:
            raise ValidationError(
                _('You can only assign a choice group to multiple choice attributes')
            )

    def get_choices(self):
        """
        Returns a query set of :class:`EnumValue` objects for this attribute.
        Returns None if the datatype of this attribute is not *TYPE_ENUM*.
        """
        return (
            self.enum_group.values.all()
            if self.datatype == Attribute.TYPE_ENUM
            else None
        )

    def save_value(self, entity, value):
        """
        Called with *entity*, any Django object registered with eav, and
        *value*, the :class:`Value` this attribute for *entity* should
        be set to.

        If a :class:`Value` object for this *entity* and attribute doesn't
        exist, one will be created.

        .. note::
           If *value* is None and a :class:`Value` object exists for this
           Attribute and *entity*, it will delete that :class:`Value` object.
        """
        ct = ContentType.objects.get_for_model(entity)

        entity_filter = {
            'entity_ct': ct,
            'attribute': self,
            '{0}'.format(get_entity_pk_type(entity)): entity.pk,
        }

        try:
            value_obj = self.value_set.get(**entity_filter)
        except Value.DoesNotExist:
            if value == None or value == '':
                return

            value_obj = Value.objects.create(**entity_filter)

        if value == None or value == '':
            value_obj.delete()
            return

        if value != value_obj.value:
            value_obj.value = value
            value_obj.save()

    def __str__(self):
        return '{} ({})'.format(self.name, self.get_datatype_display())


class Value(models.Model):  # noqa: WPS110
    """Putting the **V** in *EAV*.

    This model stores the value for one particular :class:`Attribute` for
    some entity.

    As with most EAV implementations, most of the columns of this model will
    be blank, as onle one *value_* field will be used.

    Example::

        import eav
        from django.contrib.auth.models import User

        eav.register(User)

        u = User.objects.create(username='crazy_dev_user')
        a = Attribute.objects.create(name='Fav Drink', datatype='text')

        Value.objects.create(entity = u, attribute = a, value_text = 'red bull')
        # = <Value: crazy_dev_user - Fav Drink: "red bull">
    """

    class Meta:
        verbose_name = _('Value')
        verbose_name_plural = _('Values')

    # Direct foreign keys
    attribute = models.ForeignKey(
        Attribute,
        db_index=True,
        on_delete=models.PROTECT,
        verbose_name=_('Attribute'),
    )

    # Entity generic relationships. Rather than rely on database casting,
    # this will instead use a separate ForeignKey field attribute that matches
    # the FK type of the entity.
    entity_id = models.IntegerField(
        blank=True,
        null=True,
        verbose_name=_('Entity id'),
    )

    entity_uuid = models.UUIDField(
        blank=True,
        null=True,
        verbose_name=_('Entity uuid'),
    )

    entity_ct = models.ForeignKey(
        ContentType,
        on_delete=models.PROTECT,
        related_name='value_entities',
        verbose_name=_('Entity ct'),
    )

    entity_pk_int = generic.GenericForeignKey(
        ct_field='entity_ct',
        fk_field='entity_id',
    )

    entity_pk_uuid = generic.GenericForeignKey(
        ct_field='entity_ct',
        fk_field='entity_uuid',
    )

    # Model attributes
    created = models.DateTimeField(
        default=timezone.now,
        verbose_name=_('Created'),
    )

    modified = models.DateTimeField(
        auto_now=True,
        verbose_name=_('Modified'),
    )

    # Value attributes
    value_bool = models.BooleanField(
        blank=True,
        null=True,
        verbose_name=_('Value bool'),
    )
    value_csv = CSVField(
        blank=True,
        null=True,
        verbose_name=_('Value CSV'),
    )
    value_date = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name=_('Value date'),
    )
    value_float = models.FloatField(
        blank=True,
        null=True,
        verbose_name=_('Value float'),
    )
    value_int = models.BigIntegerField(
        blank=True,
        null=True,
        verbose_name=_('Value int'),
    )
    value_text = models.TextField(
        blank=True,
        null=True,
        verbose_name=_('Value text'),
    )

    value_json = models.JSONField(
        default=dict,
        encoder=DjangoJSONEncoder,
        blank=True,
        null=True,
        verbose_name=_('Value JSON'),
    )

    value_enum = models.ForeignKey(
        EnumValue,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name='eav_values',
        verbose_name=_('Value enum'),
    )

    # Value object relationship
    generic_value_id = models.IntegerField(
        blank=True,
        null=True,
        verbose_name=_('Generic value id'),
    )

    generic_value_ct = models.ForeignKey(
        ContentType,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name='value_values',
        verbose_name=_('Generic value content type'),
    )

    value_object = generic.GenericForeignKey(
        ct_field='generic_value_ct',
        fk_field='generic_value_id',
    )

    def __str__(self):
        """String representation of a Value."""
        entity = self.entity_pk_int
        if self.entity_uuid:
            entity = self.entity_pk_uuid
        return '{0}: "{1}" ({2})'.format(
            self.attribute.name,
            self.value,
            entity,
        )

    def __repr__(self):
        """Representation of Value object."""
        entity = self.entity_pk_int
        if self.entity_uuid:
            entity = self.entity_pk_uuid
        return '{0}: "{1}" ({2})'.format(
            self.attribute.name,
            self.value,
            entity.pk,
        )

    def save(self, *args, **kwargs):
        """Validate and save this value."""
        self.full_clean()
        super().save(*args, **kwargs)

    def _get_value(self):
        """Return the python object this value is holding."""
        return getattr(self, 'value_{0}'.format(self.attribute.datatype))

    def _set_value(self, new_value):
        """Set the object this value is holding."""
        setattr(self, 'value_{0}'.format(self.attribute.datatype), new_value)

    value = property(_get_value, _set_value)  # noqa: WPS110


class Entity(object):
    """
    The helper class that will be attached to any entity
    registered with eav.
    """

    @staticmethod
    def pre_save_handler(sender, *args, **kwargs):
        """
        Pre save handler attached to self.instance.  Called before the
        model instance we are attached to is saved. This allows us to call
        :meth:`validate_attributes` before the entity is saved.
        """
        instance = kwargs['instance']
        entity = getattr(kwargs['instance'], instance._eav_config_cls.eav_attr)
        entity.validate_attributes()

    @staticmethod
    def post_save_handler(sender, *args, **kwargs):
        """
        Post save handler attached to self.instance.  Calls :meth:`save` when
        the model instance we are attached to is saved.
        """
        instance = kwargs['instance']
        entity = getattr(instance, instance._eav_config_cls.eav_attr)
        entity.save()

    def __init__(self, instance):
        """
        Set self.instance equal to the instance of the model that we're attached
        to. Also, store the content type of that instance.
        """
        self.instance = instance
        self.ct = ContentType.objects.get_for_model(instance)

    def __getattr__(self, name):
        """
        Tha magic getattr helper. This is called whenever user invokes::

            instance.<attribute>

        Checks if *name* is a valid slug for attributes available to this
        instances. If it is, tries to lookup the :class:`Value` with that
        attribute slug. If there is one, it returns the value of the
        class:`Value` object, otherwise it hasn't been set, so it returns
        None.
        """
        if not name.startswith('_'):
            try:
                attribute = self.get_attribute_by_slug(name)
            except Attribute.DoesNotExist:
                raise AttributeError(
                    _('%(obj)s has no EAV attribute named %(attr)s')
                    % dict(obj=self.instance, attr=name)
                )

            try:
                return self.get_value_by_attribute(attribute).value
            except Value.DoesNotExist:
                return None

        return getattr(super(Entity, self), name)

    def get_all_attributes(self):
        """
        Return a query set of all :class:`Attribute` objects that can be set
        for this entity.
        """
        return self.instance._eav_config_cls.get_attributes(
            instance=self.instance
        ).order_by('display_order')

    def _hasattr(self, attribute_slug):
        """
        Since we override __getattr__ with a backdown to the database, this
        exists as a way of checking whether a user has set a real attribute on
        ourselves, without going to the db if not.
        """
        return attribute_slug in self.__dict__

    def _getattr(self, attribute_slug):
        """
        Since we override __getattr__ with a backdown to the database, this
        exists as a way of getting the value a user set for one of our
        attributes, without going to the db to check.
        """
        return self.__dict__[attribute_slug]

    def save(self):
        """
        Saves all the EAV values that have been set on this entity.
        """
        for attribute in self.get_all_attributes():
            if self._hasattr(attribute.slug):
                attribute_value = self._getattr(attribute.slug)
                if attribute.datatype == Attribute.TYPE_ENUM and not isinstance(
                    attribute_value, EnumValue
                ):
                    if attribute_value is not None:
                        attribute_value = EnumValue.objects.get(value=attribute_value)
                attribute.save_value(self.instance, attribute_value)

    def validate_attributes(self):
        """
        Called before :meth:`save`, first validate all the entity values to
        make sure they can be created / saved cleanly.
        Raises ``ValidationError`` if they can't be.
        """
        values_dict = self.get_values_dict()

        for attribute in self.get_all_attributes():
            value = None

            # Value was assigned to this instance.
            if self._hasattr(attribute.slug):
                value = self._getattr(attribute.slug)
                values_dict.pop(attribute.slug, None)
            # Otherwise try pre-loaded from DB.
            else:
                value = values_dict.pop(attribute.slug, None)

            if value is None:
                if attribute.required:
                    raise ValidationError(
                        _('{} EAV field cannot be blank'.format(attribute.slug))
                    )
            else:
                try:
                    attribute.validate_value(value)
                except ValidationError as e:
                    raise ValidationError(
                        _('%(attr)s EAV field %(err)s')
                        % dict(attr=attribute.slug, err=e)
                    )

        illegal = values_dict or (
            self.get_object_attributes() - self.get_all_attribute_slugs()
        )

        if illegal:
            raise IllegalAssignmentException(
                'Instance of the class {} cannot have values for attributes: {}.'.format(
                    self.instance.__class__, ', '.join(illegal)
                )
            )

    def get_values_dict(self):
        return {v.attribute.slug: v.value for v in self.get_values()}

    def get_values(self):
        """Get all set :class:`Value` objects for self.instance."""
        entity_filter = {
            'entity_ct': self.ct,
            '{0}'.format(get_entity_pk_type(self.instance)): self.instance.pk,
        }

        return Value.objects.filter(**entity_filter).select_related()

    def get_all_attribute_slugs(self):
        """
        Returns a list of slugs for all attributes available to this entity.
        """
        return set(self.get_all_attributes().values_list('slug', flat=True))

    def get_attribute_by_slug(self, slug):
        """
        Returns a single :class:`Attribute` with *slug*.
        """
        return self.get_all_attributes().get(slug=slug)

    def get_value_by_attribute(self, attribute):
        """
        Returns a single :class:`Value` for *attribute*.
        """
        return self.get_values().get(attribute=attribute)

    def get_object_attributes(self):
        """
        Returns entity instance attributes, except for
        ``instance`` and ``ct`` which are used internally.
        """
        return set(copy(self.__dict__).keys()) - set(['instance', 'ct'])

    def __iter__(self):
        """
        Iterate over set eav values. This would allow you to do::

            for i in m.eav: print(i)
        """
        return iter(self.get_values())


class EAVModelMeta(ModelBase):
    def __new__(cls, name, bases, namespace, **kwds):
        result = super(EAVModelMeta, cls).__new__(cls, name, bases, dict(namespace))
        register(result)
        return result



from django.db import migrations, models


class Migration(migrations.Migration):
    """Convert Value.value_int to BigInteger."""

    dependencies = [
        ('eav', '0006_add_entity_uuid'),
    ]

    operations = [
        migrations.AlterField(
            model_name='value',
            name='value_int',
            field=models.BigIntegerField(blank=True, null=True),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    """Creates UUID field to map to Entity FK."""

    dependencies = [
        ('eav', '0005_auto_20210510_1305'),
    ]

    operations = [
        migrations.AddField(
            model_name='value',
            name='entity_uuid',
            field=models.UUIDField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='value',
            name='entity_id',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    """Use Django SlugField() for Attribute.slug."""

    dependencies = [
        ('eav', '0007_alter_value_value_int'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attribute',
            name='slug',
            field=models.SlugField(
                help_text='Short unique attribute label',
                unique=True,
                verbose_name='Slug',
            ),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    """Add entity_ct field to Attribute model."""

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('eav', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='entity_ct',
            field=models.ManyToManyField(blank=True, to='contenttypes.ContentType'),
        ),
    ]



"""This module contains classes used for admin integration."""

from django.contrib import admin
from django.contrib.admin.options import InlineModelAdmin, ModelAdmin
from django.forms.models import BaseInlineFormSet
from django.utils.safestring import mark_safe

from eav.models import Attribute, EnumGroup, EnumValue, Value


class BaseEntityAdmin(ModelAdmin):
    def render_change_form(self, request, context, *args, **kwargs):
        """
        Wrapper for ``ModelAdmin.render_change_form``. Replaces standard static
        ``AdminForm`` with an EAV-friendly one. The point is that our form
        generates fields dynamically and fieldsets must be inferred from a
        prepared and validated form instance, not just the form class. Django
        does not seem to provide hooks for this purpose, so we simply wrap the
        view and substitute some data.
        """
        form = context['adminform'].form
        media = context["media"]

        # Infer correct data from the form.
        fieldsets = self.fieldsets or [(None, {'fields': form.fields.keys()})]
        adminform = admin.helpers.AdminForm(form, fieldsets, self.prepopulated_fields)
        media = mark_safe(media + adminform.media)

        context.update(adminform=adminform, media=media)

        return super(BaseEntityAdmin, self).render_change_form(
            request, context, *args, **kwargs
        )


class BaseEntityInlineFormSet(BaseInlineFormSet):
    """
    An inline formset that correctly initializes EAV forms.
    """

    def add_fields(self, form, index):
        if self.instance:
            setattr(form.instance, self.fk.name, self.instance)
            form._build_dynamic_fields()

        super(BaseEntityInlineFormSet, self).add_fields(form, index)


class BaseEntityInline(InlineModelAdmin):
    """
    Inline model admin that works correctly with EAV attributes. You should mix
    in the standard ``StackedInline`` or ``TabularInline`` classes in order to
    define formset representation, e.g.::

        class ItemInline(BaseEntityInline, StackedInline):
            model = Item
            form = forms.ItemForm

    .. warning:: ``TabularInline`` does *not* work out of the box. There is,
        however, a patched template ``admin/edit_inline/tabular.html`` bundled
        with EAV-Django. You can copy or symlink the ``admin`` directory to
        your templates search path (see Django documentation).
    """

    formset = BaseEntityInlineFormSet

    def get_fieldsets(self, request, obj=None):
        if self.declared_fieldsets:
            return self.declared_fieldsets

        formset = self.get_formset(request)
        fk_name = self.fk_name or formset.fk.name
        kw = {fk_name: obj} if obj else {}
        instance = self.model(**kw)
        form = formset.form(request.POST, instance=instance)

        return [(None, {'fields': form.fields.keys()})]


class AttributeAdmin(ModelAdmin):
    list_display = ('name', 'slug', 'datatype', 'description')
    prepopulated_fields = {'slug': ('name',)}


admin.site.register(Attribute, AttributeAdmin)
admin.site.register(EnumValue)
admin.site.register(EnumGroup)
admin.site.register(Value)



"""
This module contains pure wrapper functions used as decorators.
Functions in this module should be simple and not involve complex logic.
"""


def register_eav(**kwargs):
    """
    Registers the given model(s) classes and wrapped ``Model`` class with
    Django EAV 2::

        @register_eav
        class Author(models.Model):
            pass
    """
    from django.db.models import Model

    from eav import register

    def _model_eav_wrapper(model_class):
        if not issubclass(model_class, Model):
            raise ValueError('Wrapped class must subclass Model.')
        register(model_class, **kwargs)
        return model_class

    return _model_eav_wrapper



# Generated by Django 3.1.6 on 2021-04-04 22:09

import django.core.serializers.json
from django.db import migrations
from django.db.models import JSONField

import eav.fields


class Migration(migrations.Migration):
    dependencies = [
        ('eav', '0002_add_entity_ct_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='value',
            name='value_json',
            field=JSONField(
                blank=True,
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='attribute',
            name='datatype',
            field=eav.fields.EavDatatypeField(
                choices=[
                    ('text', 'Text'),
                    ('date', 'Date'),
                    ('float', 'Float'),
                    ('int', 'Integer'),
                    ('bool', 'True / False'),
                    ('object', 'Django Object'),
                    ('enum', 'Multiple Choice'),
                    ('json', 'JSON Object'),
                ],
                max_length=6,
                verbose_name='Data Type',
            ),
        ),
    ]



"""
This module contains the custom manager used by entities registered with eav.
"""

from django.db import models

from eav.queryset import EavQuerySet


class EntityManager(models.Manager):
    """
    Our custom manager, overrides ``models.Manager``.
    """

    _queryset_class = EavQuerySet

    def create(self, **kwargs):
        """
        Parse eav attributes out of *kwargs*, then try to create and save
        the object, then assign and save it's eav attributes.
        """
        config_cls = getattr(self.model, '_eav_config_cls', None)

        if not config_cls or config_cls.manager_only:
            return super(EntityManager, self).create(**kwargs)

        prefix = '%s__' % config_cls.eav_attr
        new_kwargs = {}
        eav_kwargs = {}

        for key, value in kwargs.items():
            if key.startswith(prefix):
                eav_kwargs.update({key[len(prefix) :]: value})
            else:
                new_kwargs.update({key: value})

        obj = self.model(**new_kwargs)
        obj_eav = getattr(obj, config_cls.eav_attr)

        for key, value in eav_kwargs.items():
            setattr(obj_eav, key, value)

        obj.save()
        return obj

    def get_or_create(self, defaults=None, **kwargs):
        """
        Reproduces the behavior of get_or_create, eav friendly.
        """
        try:
            return self.get(**kwargs), False
        except self.model.DoesNotExist:
            if defaults:
                kwargs = {**kwargs, **defaults}
            return self.create(**kwargs), True



def register(model_cls, config_cls=None):
    from eav.registry import Registry

    Registry.register(model_cls, config_cls)


def unregister(model_cls):
    from eav.registry import Registry

    Registry.unregister(model_cls)



"""This modules contains the registry classes."""

from django.contrib.contenttypes import fields as generic
from django.db.models.signals import post_init, post_save, pre_save

from eav.managers import EntityManager
from eav.models import Attribute, Entity, Value

from eav.logic.entity_pk import get_entity_pk_type


class EavConfig(object):
    """
    The default ``EavConfig`` class used if it is not overriden on registration.
    This is where all the default eav attribute names are defined.

    Available options are as follows:

    1. manager_attr - Specifies manager name. Used to refer to the
       manager from Entity class, "objects" by default.
    2. manager_only - Specifies whether signals and generic relation should
       be setup for the registered model.
    3. eav_attr - Named of the Entity toolkit instance on the registered
       model instance. "eav" by default. See attach_eav_attr.
    4. generic_relation_attr - Name of the GenericRelation to Value
       objects. "eav_values" by default.
    5. generic_relation_related_name - Name of the related name for
       GenericRelation from Entity to Value. None by default. Therefore,
       if not overridden, it is not possible to query Values by Entities.
    """

    manager_attr = 'objects'
    manager_only = False
    eav_attr = 'eav'
    generic_relation_attr = 'eav_values'
    generic_relation_related_name = None

    @classmethod
    def get_attributes(cls, instance=None):
        """
        By default, all :class:`~eav.models.Attribute` object apply to an
        entity, unless you provide a custom EavConfig class overriding this.
        """
        return Attribute.objects.all()


class Registry(object):
    """
    Handles registration through the
    :meth:`register` and :meth:`unregister` methods.
    """

    @staticmethod
    def register(model_cls, config_cls=None):
        """
        Registers *model_cls* with eav. You can pass an optional *config_cls*
        to override the EavConfig defaults.

        .. note::
           Multiple registrations for the same entity are harmlessly ignored.
        """
        if hasattr(model_cls, '_eav_config_cls'):
            return

        if config_cls is EavConfig or config_cls is None:
            config_cls = type("%sConfig" % model_cls.__name__, (EavConfig,), {})

        # set _eav_config_cls on the model so we can access it there
        setattr(model_cls, '_eav_config_cls', config_cls)

        reg = Registry(model_cls)
        reg._register_self()

    @staticmethod
    def unregister(model_cls):
        """
        Unregisters *model_cls* with eav.

        .. note::
           Unregistering a class not already registered is harmlessly ignored.
        """
        if not getattr(model_cls, '_eav_config_cls', None):
            return
        reg = Registry(model_cls)
        reg._unregister_self()

        delattr(model_cls, '_eav_config_cls')

    @staticmethod
    def attach_eav_attr(sender, *args, **kwargs):
        """
        Attach EAV Entity toolkit to an instance after init.
        """
        instance = kwargs['instance']
        config_cls = instance.__class__._eav_config_cls
        setattr(instance, config_cls.eav_attr, Entity(instance))

    def __init__(self, model_cls):
        """
        Set the *model_cls* and its *config_cls*
        """
        self.model_cls = model_cls
        self.config_cls = model_cls._eav_config_cls

    def _attach_manager(self):
        """
        Attach the manager to *manager_attr* specified in *config_cls*
        """
        # Save the old manager if the attribute name conflicts with the new one.
        if hasattr(self.model_cls, self.config_cls.manager_attr):
            mgr = getattr(self.model_cls, self.config_cls.manager_attr)

            # For some models, `local_managers` may be empty, eg.
            # django.contrib.auth.models.User and AbstractUser
            if mgr in self.model_cls._meta.local_managers:
                self.config_cls.old_mgr = mgr
                self.model_cls._meta.local_managers.remove(mgr)

            self.model_cls._meta._expire_cache()

        # Attach the new manager to the model.
        mgr = EntityManager()
        mgr.contribute_to_class(self.model_cls, self.config_cls.manager_attr)

    def _detach_manager(self):
        """
        Detach the manager and restore the previous one (if there was one).
        """
        mgr = getattr(self.model_cls, self.config_cls.manager_attr)
        self.model_cls._meta.local_managers.remove(mgr)
        self.model_cls._meta._expire_cache()
        delattr(self.model_cls, self.config_cls.manager_attr)

        if hasattr(self.config_cls, 'old_mgr'):
            self.config_cls.old_mgr.contribute_to_class(
                self.model_cls, self.config_cls.manager_attr
            )

    def _attach_signals(self):
        """
        Attach pre- and post- save signals from model class
        to Entity helper. This way, Entity instance will be
        able to prepare and clean-up before and after creation /
        update of the user's model class instance.
        """
        post_init.connect(Registry.attach_eav_attr, sender=self.model_cls)
        pre_save.connect(Entity.pre_save_handler, sender=self.model_cls)
        post_save.connect(Entity.post_save_handler, sender=self.model_cls)

    def _detach_signals(self):
        """
        Detach all signals for eav.
        """
        post_init.disconnect(Registry.attach_eav_attr, sender=self.model_cls)
        pre_save.disconnect(Entity.pre_save_handler, sender=self.model_cls)
        post_save.disconnect(Entity.post_save_handler, sender=self.model_cls)

    def _attach_generic_relation(self):
        """Set up the generic relation for the entity."""
        rel_name = (
            self.config_cls.generic_relation_related_name or self.model_cls.__name__
        )

        gr_name = self.config_cls.generic_relation_attr.lower()
        generic_relation = generic.GenericRelation(
            Value,
            object_id_field=get_entity_pk_type(self.model_cls),
            content_type_field='entity_ct',
            related_query_name=rel_name,
        )
        generic_relation.contribute_to_class(self.model_cls, gr_name)

    def _detach_generic_relation(self):
        """
        Remove the generic relation from the entity
        """
        gen_rel_field = self.config_cls.generic_relation_attr.lower()
        for field in self.model_cls._meta.local_many_to_many:
            if field.name == gen_rel_field:
                self.model_cls._meta.local_many_to_many.remove(field)
                break

        delattr(self.model_cls, gen_rel_field)

    def _register_self(self):
        """
        Call the necessary registration methods
        """
        self._attach_manager()

        if not self.config_cls.manager_only:
            self._attach_signals()
            self._attach_generic_relation()

    def _unregister_self(self):
        """
        Call the necessary unregistration methods
        """
        self._detach_manager()

        if not self.config_cls.manager_only:
            self._detach_signals()
            self._detach_generic_relation()



import secrets
import string

from django.utils.text import slugify

try:
    from typing import Final
except ImportError:
    from typing_extensions import Final

SLUGFIELD_MAX_LENGTH: Final = 50


def generate_slug(name: str) -> str:
    """Generates a valid slug based on ``name``."""
    slug = slugify(name, allow_unicode=False)

    if not slug:
        # Fallback to ensure a slug is always generated by using a random one
        chars = string.ascii_lowercase + string.digits
        randstr = ''.join(secrets.choice(chars) for _ in range(8))
        slug = 'rand-{0}'.format(randstr)

    slug = slug.encode('utf-8', 'surrogateescape').decode()

    return slug[:SLUGFIELD_MAX_LENGTH]



# Generated by Django 2.0.4 on 2018-06-01 09:36

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models

import eav.fields


class Migration(migrations.Migration):
    """Initial migration that creates the Attribute, EnumGroup, EnumValue, and Value models."""

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attribute',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(
                        help_text='User-friendly attribute name',
                        max_length=100,
                        verbose_name='Name',
                    ),
                ),
                (
                    'slug',
                    models.SlugField(
                        help_text='Short unique attribute label',
                        unique=True,
                        verbose_name='Slug',
                    ),
                ),
                (
                    'description',
                    models.CharField(
                        blank=True,
                        help_text='Short description',
                        max_length=256,
                        null=True,
                        verbose_name='Description',
                    ),
                ),
                (
                    'datatype',
                    eav.fields.EavDatatypeField(
                        choices=[
                            ('text', 'Text'),
                            ('date', 'Date'),
                            ('float', 'Float'),
                            ('int', 'Integer'),
                            ('bool', 'True / False'),
                            ('object', 'Django Object'),
                            ('enum', 'Multiple Choice'),
                        ],
                        max_length=6,
                        verbose_name='Data Type',
                    ),
                ),
                (
                    'created',
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name='Created',
                    ),
                ),
                (
                    'modified',
                    models.DateTimeField(auto_now=True, verbose_name='Modified'),
                ),
                (
                    'required',
                    models.BooleanField(default=False, verbose_name='Required'),
                ),
                (
                    'display_order',
                    models.PositiveIntegerField(
                        default=1, verbose_name='Display order'
                    ),
                ),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='EnumGroup',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'name',
                    models.CharField(max_length=100, unique=True, verbose_name='Name'),
                ),
            ],
        ),
        migrations.CreateModel(
            name='EnumValue',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'value',
                    models.CharField(
                        db_index=True, max_length=50, unique=True, verbose_name='Value'
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='Value',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('entity_id', models.IntegerField()),
                ('value_text', models.TextField(blank=True, null=True)),
                ('value_float', models.FloatField(blank=True, null=True)),
                ('value_int', models.IntegerField(blank=True, null=True)),
                ('value_date', models.DateTimeField(blank=True, null=True)),
                ('value_bool', models.NullBooleanField()),
                ('generic_value_id', models.IntegerField(blank=True, null=True)),
                (
                    'created',
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name='Created'
                    ),
                ),
                (
                    'modified',
                    models.DateTimeField(auto_now=True, verbose_name='Modified'),
                ),
                (
                    'attribute',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to='eav.Attribute',
                        verbose_name='Attribute',
                    ),
                ),
                (
                    'entity_ct',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='value_entities',
                        to='contenttypes.ContentType',
                    ),
                ),
                (
                    'generic_value_ct',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='value_values',
                        to='contenttypes.ContentType',
                    ),
                ),
                (
                    'value_enum',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='eav_values',
                        to='eav.EnumValue',
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name='enumgroup',
            name='values',
            field=models.ManyToManyField(to='eav.EnumValue', verbose_name='Enum group'),
        ),
        migrations.AddField(
            model_name='attribute',
            name='enum_group',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to='eav.EnumGroup',
                verbose_name='Choice Group',
            ),
        ),
    ]



from django.core.serializers.json import DjangoJSONEncoder
from django.db import migrations, models

from eav.fields import CSVField


class Migration(migrations.Migration):
    """Define verbose naming for models and fields."""

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('eav', '0008_use_native_slugfield'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='attribute',
            options={
                'ordering': ['name'],
                'verbose_name': 'Attribute',
                'verbose_name_plural': 'Attributes',
            },
        ),
        migrations.AlterModelOptions(
            name='enumgroup',
            options={
                'verbose_name': 'EnumGroup',
                'verbose_name_plural': 'EnumGroups',
            },
        ),
        migrations.AlterModelOptions(
            name='enumvalue',
            options={
                'verbose_name': 'EnumValue',
                'verbose_name_plural': 'EnumValues',
            },
        ),
        migrations.AlterModelOptions(
            name='value',
            options={'verbose_name': 'Value', 'verbose_name_plural': 'Values'},
        ),
        migrations.AlterField(
            model_name='attribute',
            name='entity_ct',
            field=models.ManyToManyField(
                blank=True,
                to='contenttypes.contenttype',
                verbose_name='Entity content type',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='entity_ct',
            field=models.ForeignKey(
                on_delete=models.deletion.PROTECT,
                related_name='value_entities',
                to='contenttypes.contenttype',
                verbose_name='Entity ct',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='entity_id',
            field=models.IntegerField(
                blank=True,
                null=True,
                verbose_name='Entity id',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='entity_uuid',
            field=models.UUIDField(
                blank=True,
                null=True,
                verbose_name='Entity uuid',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='generic_value_ct',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.deletion.PROTECT,
                related_name='value_values',
                to='contenttypes.contenttype',
                verbose_name='Generic value content type',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='generic_value_id',
            field=models.IntegerField(
                blank=True,
                null=True,
                verbose_name='Generic value id',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_bool',
            field=models.BooleanField(
                blank=True,
                null=True,
                verbose_name='Value bool',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_csv',
            field=CSVField(
                blank=True,
                default='',
                null=True,
                verbose_name='Value CSV',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_date',
            field=models.DateTimeField(
                blank=True,
                null=True,
                verbose_name='Value date',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_enum',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.deletion.PROTECT,
                related_name='eav_values',
                to='eav.enumvalue',
                verbose_name='Value enum',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_float',
            field=models.FloatField(
                blank=True,
                null=True,
                verbose_name='Value float',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_int',
            field=models.BigIntegerField(
                blank=True,
                null=True,
                verbose_name='Value int',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_json',
            field=models.JSONField(
                blank=True,
                default=dict,
                encoder=DjangoJSONEncoder,
                null=True,
                verbose_name='Value JSON',
            ),
        ),
        migrations.AlterField(
            model_name='value',
            name='value_text',
            field=models.TextField(
                blank=True,
                null=True,
                verbose_name='Value text',
            ),
        ),
    ]



# Generated by Django 3.2 on 2021-04-23 19:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('eav', '0003_auto_20210404_2209'),
    ]

    operations = [
        migrations.AlterField(
            model_name='value',
            name='value_bool',
            field=models.BooleanField(blank=True, null=True),
        ),
    ]



from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from eav.forms import CSVFormField


class EavDatatypeField(models.CharField):
    """
    The datatype field used by :class:`~eav.models.Attribute`.
    """

    def validate(self, value, instance):
        """
        Raise ``ValidationError`` if they try to change the datatype of an
        :class:`~eav.models.Attribute` that is already used by
        :class:`~eav.models.Value` objects.
        """
        super(EavDatatypeField, self).validate(value, instance)

        if not instance.pk:
            return

        if type(instance).objects.get(pk=instance.pk).datatype == instance.datatype:
            return

        if instance.value_set.count():
            raise ValidationError(
                _(
                    'You cannot change the datatype of an attribute that is already in use.'
                )
            )


class CSVField(models.TextField):  # (models.Field):
    description = _("A Comma-Separated-Value field.")
    default_separator = ";"

    def __init__(self, separator=";", *args, **kwargs):
        self.separator = separator
        kwargs.setdefault('default', "")
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.separator != self.default_separator:
            kwargs['separator'] = self.separator
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        defaults = {'form_class': CSVFormField}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    def from_db_value(self, value, expression, connection, context=None):
        if value is None:
            return []
        return value.split(self.separator)

    def to_python(self, value):
        if value is None:
            return []
        if isinstance(value, list):
            return value
        return value.split(self.separator)

    def get_prep_value(self, value):
        if not value:
            return ""
        if isinstance(value, str):
            return value
        elif isinstance(value, list):
            return self.separator.join(value)

    def value_to_string(self, obj):
        value = self.value_from_object(obj)
        return self.get_prep_value(value)
